package de.mdelab.mlsdm.interpreter.patternMatcher.patternPartBased;

import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.LinkOrderConstraint;
import de.mdelab.mlstorypatterns.LinkOrderConstraintEnum;
import de.mdelab.mlstorypatterns.LinkPositionConstraintEnum;
import de.mdelab.mlstorypatterns.MatchTypeEnum;
import de.mdelab.mlstorypatterns.ModifierEnum;
import de.mdelab.mlstorypatterns.StoryPatternLink;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.CheckPatternPartTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.ECheckResult;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.EMatchType;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.MatchState;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.MatchTransaction;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPart;
import de.mdelab.sdm.interpreter.core.patternmatcher.patternPartBased.PatternPartBasedMatcher;
import de.mdelab.sdm.interpreter.core.variables.Variable;

/**
 * 
 * @author Stephan Hildebrandt
 * 
 */
public class MLSDMStoryPatternLinkPatternPart extends MLSDMPatternPart<AbstractStoryPatternObject, StoryPatternLink>
{
	public MLSDMStoryPatternLinkPatternPart(
			final PatternPartBasedMatcher<?, ?, ?, ?, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, ?, MLExpression> patternMatcher,
			final StoryPatternLink storyPatternLink)
	{
		super(patternMatcher, storyPatternLink, new AbstractStoryPatternObject[]
		{
				storyPatternLink.getSource(), storyPatternLink.getTarget()
		});

		/*
		 * if a story pattern object is created, the link must also be created
		 */
		assert storyPatternLink.getSource().getModifier() != ModifierEnum.CREATE || storyPatternLink.getModifier() == ModifierEnum.CREATE;
		assert storyPatternLink.getTarget().getModifier() != ModifierEnum.CREATE || storyPatternLink.getModifier() == ModifierEnum.CREATE;

		/*
		 * if a story pattern object is destroyed, the link must also be
		 * destroyed
		 */
		assert storyPatternLink.getSource().getModifier() != ModifierEnum.DESTROY || storyPatternLink.getModifier() == ModifierEnum.DESTROY;
		assert storyPatternLink.getTarget().getModifier() != ModifierEnum.DESTROY || storyPatternLink.getModifier() == ModifierEnum.DESTROY;

		/*
		 * If one of the link ends is optional, the links must be optional as
		 * well.
		 */
		assert storyPatternLink.getSource().getMatchType() != MatchTypeEnum.OPTIONAL
				|| storyPatternLink.getMatchType() == MatchTypeEnum.OPTIONAL;
		assert storyPatternLink.getTarget().getMatchType() != MatchTypeEnum.OPTIONAL
				|| storyPatternLink.getMatchType() == MatchTypeEnum.OPTIONAL;

		/*
		 * If the link is created or destroyed, the link's feature must be
		 * changeable
		 */
		assert storyPatternLink.getModifier() != ModifierEnum.CREATE && storyPatternLink.getModifier() != ModifierEnum.DESTROY
				|| storyPatternLink.getFeature().isChangeable();
	}

	@Override
	public MatchState createMatchState()
	{
		return new MLSDMStoryPatternLinkMatchState<Iterator<Object>>();
	}

	@SuppressWarnings("unchecked")
	@Override
	public ECheckResult check() throws SDMException
	{
		final AbstractStoryPatternObject source = this.storyPatternLink.getSource();
		final AbstractStoryPatternObject target = this.storyPatternLink.getTarget();

		if (!this.isCreate && this.patternMatcher.isBound(source) && this.patternMatcher.isBound(target))
		{
			final EObject sourceInstanceObject = (EObject) this.patternMatcher.getInstanceObject(source);
			final Object targetInstanceObject = this.patternMatcher.getInstanceObject(target);

			final EStructuralFeature eStructuralFeature = this.storyPatternLink.getFeature();

			if (!eStructuralFeature.isMany())
			{
				assert this.storyPatternLink.getLinkPositionConstraint() == LinkPositionConstraintEnum.UNCONSTRAINED;
				assert this.storyPatternLink.getIncomingLinkOrderConstraints().isEmpty();
				assert this.storyPatternLink.getOutgoingLinkOrderConstraints().isEmpty();

				if (sourceInstanceObject.eGet(eStructuralFeature) == targetInstanceObject)
				{
					this.patternMatcher.getNotificationEmitter().linkCheckSuccessful(source, sourceInstanceObject, this.storyPatternLink,
							target, targetInstanceObject, this.patternMatcher.getVariablesScope(), this.patternMatcher);

					return ECheckResult.OK;
				}
				else
				{
					this.patternMatcher.getNotificationEmitter().linkCheckFailed(source, sourceInstanceObject, this.storyPatternLink,
							target, targetInstanceObject, this.patternMatcher.getVariablesScope(), this.patternMatcher);

					return ECheckResult.FAIL;
				}
			}
			else
			{
				if (((Collection<Object>) sourceInstanceObject.eGet(eStructuralFeature)).contains(targetInstanceObject))
				{
					this.patternMatcher.getNotificationEmitter().linkCheckSuccessful(source, sourceInstanceObject, this.storyPatternLink,
							target, targetInstanceObject, this.patternMatcher.getVariablesScope(), this.patternMatcher);

					return ECheckResult.OK;
				}
				else
				{
					this.patternMatcher.getNotificationEmitter().linkCheckFailed(source, sourceInstanceObject, this.storyPatternLink,
							target, targetInstanceObject, this.patternMatcher.getVariablesScope(), this.patternMatcher);

					return ECheckResult.FAIL;
				}
			}
		}
		else if (this.isCreate
				&& (this.storyPatternLink.getMatchType() != MatchTypeEnum.OPTIONAL || this.patternMatcher.isBound(this.storyPatternLink
						.getSource()) && this.patternMatcher.isBound(this.storyPatternLink.getTarget())))
		{
			return ECheckResult.OK;
		}
		else
		{
			return ECheckResult.UNKNOWN;
		}

	}

	@SuppressWarnings("unchecked")
	@Override
	public boolean match(final MatchState matchState) throws SDMException
	{
		assert matchState instanceof MLSDMStoryPatternLinkMatchState;

		final MLSDMStoryPatternLinkMatchState<Iterator<Object>> ms = (MLSDMStoryPatternLinkMatchState<Iterator<Object>>) matchState;

		AbstractStoryPatternObject sourceSpo = this.storyPatternLink.getSource();
		AbstractStoryPatternObject targetSpo = this.storyPatternLink.getTarget();

		assert this.patternMatcher.isBound(sourceSpo) || this.patternMatcher.isBound(targetSpo);
		assert !(this.patternMatcher.isBound(sourceSpo) && this.patternMatcher.isBound(targetSpo));

		/*
		 * The source object is bound. Search for a match of the target object.
		 */
		if (this.patternMatcher.isBound(sourceSpo))
		{
			final EObject sourceInstanceObject = (EObject) this.patternMatcher.getInstanceObject(sourceSpo);

			return this.matchTargetObject(ms, sourceSpo, targetSpo, sourceInstanceObject, this.storyPatternLink.getFeature());
		}
		else
		{
			/*
			 * Start matching from the target object
			 */
			sourceSpo = this.storyPatternLink.getTarget();
			targetSpo = this.storyPatternLink.getSource();

			final EObject sourceInstanceObject = (EObject) this.patternMatcher.getInstanceObject(sourceSpo);

			if (this.storyPatternLink.getFeature() instanceof EReference)
			{
				final EReference eReference = (EReference) this.storyPatternLink.getFeature();

				/*
				 * The link's reference has an opposite. Navigate via the
				 * opposite reference.
				 */
				if (eReference.getEOpposite() != null)
				{
					return this.matchTargetObject(ms, sourceSpo, targetSpo, sourceInstanceObject, eReference.getEOpposite());
				}
				/*
				 * The link's reference is a containment.
				 */
				else if (eReference.isContainment())
				{
					this.patternMatcher.getNotificationEmitter().traversingLink(this.storyPatternLink, sourceSpo, sourceInstanceObject,
							targetSpo, this.patternMatcher.getVariablesScope(), this.patternMatcher);

					final EObject targetInstanceObject = sourceInstanceObject.eContainer();

					if (this.patternMatcher.matchStoryPatternObject(targetSpo, targetInstanceObject)
							&& (!eReference.isMany() && targetInstanceObject.eGet(eReference) == sourceInstanceObject || eReference
									.isMany()
									&& ((Collection<Object>) targetInstanceObject.eGet(eReference)).contains(sourceInstanceObject)))
					{
						this.patternMatcher.getNotificationEmitter().storyPatternObjectBound(targetSpo, targetInstanceObject,
								this.patternMatcher.getVariablesScope(), this.patternMatcher);

						return true;
					}
				}
				else
				{
					throw new UnsupportedOperationException();
				}
			}
			else
			{
				throw new UnsupportedOperationException();
			}

			this.patternMatcher.getNotificationEmitter().storyPatternObjectNotBound(targetSpo, this.patternMatcher.getVariablesScope(),
					this.patternMatcher);
		}

		return false;
	}

	@SuppressWarnings("unchecked")
	private boolean matchTargetObject(final MLSDMStoryPatternLinkMatchState<Iterator<Object>> matchState,
			final AbstractStoryPatternObject sourceSpo, final AbstractStoryPatternObject targetSpo, final EObject sourceInstanceObject,
			final EStructuralFeature feature) throws SDMException
	{
		assert matchState != null;
		assert sourceSpo != null;
		assert targetSpo != null;
		assert sourceInstanceObject != null;
		assert feature != null;
		
		// DB ticket #1136
		assert  matchState.getSourceInstanceObject() == null || sourceInstanceObject == matchState.getSourceInstanceObject() : "Source instance object should be the same as that of match state!";

		this.patternMatcher.getNotificationEmitter().traversingLink(this.storyPatternLink, sourceSpo, sourceInstanceObject, targetSpo,
				this.patternMatcher.getVariablesScope(), this.patternMatcher);

		/*
		 * This is a to-one feature.
		 */
		if (!feature.isMany())
		{
			/*
			 * Link constraints for to-one features do not make sense.
			 */
			assert this.storyPatternLink.getLinkPositionConstraint() == LinkPositionConstraintEnum.UNCONSTRAINED;
			assert this.storyPatternLink.getOutgoingLinkOrderConstraints().isEmpty();
			assert this.storyPatternLink.getIncomingLinkOrderConstraints().isEmpty();

			final Object targetInstanceObject = sourceInstanceObject.eGet(feature);

			if (this.patternMatcher.matchStoryPatternObject(targetSpo, targetInstanceObject))
			{
				this.patternMatcher.getNotificationEmitter().storyPatternObjectBound(targetSpo, targetInstanceObject,
						this.patternMatcher.getVariablesScope(), this.patternMatcher);

				return true;
			}
		}
		/*
		 * This is a to-many feature.
		 */
		else {
			final LinkPositionConstraintEnum posCst = getStoryPatternLink().getLinkPositionConstraint();
			
			if ( posCst == null || posCst == LinkPositionConstraintEnum.UNCONSTRAINED ) {
				Iterator<Object> linkIterator = matchState.getLinkIterator();
											// DB ticket #1136: Converted into assert
				if (linkIterator == null /*|| sourceInstanceObject != matchState.getSourceInstanceObject()*/)
				{
					linkIterator = ((Collection<Object>) sourceInstanceObject.eGet(feature)).iterator();
	
					matchState.setLinkIterator(linkIterator);
					matchState.setSourceInstanceObject(sourceInstanceObject);
				}
	
				while (linkIterator.hasNext())
				{
					final Object targetInstanceObject = linkIterator.next();
	
					if (this.patternMatcher.matchStoryPatternObject(targetSpo, targetInstanceObject))
					{
						this.patternMatcher.getNotificationEmitter().storyPatternObjectBound(targetSpo, targetInstanceObject,
								this.patternMatcher.getVariablesScope(), this.patternMatcher);
	
						return true;
					}
				}
			}
			else {
				final Object matchedTargetObj = match( 	(Collection<Object>) sourceInstanceObject.eGet(feature), 
														targetSpo,
														null,
														posCst );
								
				if ( matchedTargetObj == null ) {
					return false;
				}

				patternMatcher.getNotificationEmitter().storyPatternObjectBound( 	targetSpo,
																					matchedTargetObj,
																					patternMatcher.getVariablesScope(),
																					patternMatcher );
				
				return true;
			}
		}

		this.patternMatcher.getNotificationEmitter().storyPatternObjectNotBound(targetSpo, this.patternMatcher.getVariablesScope(),
				this.patternMatcher);

		return false;
	}
		
	private Object match( 	final Collection<Object> p_objects,
							final AbstractStoryPatternObject p_targetSpo, 
							final Object p_targetObject,
							final LinkPositionConstraintEnum p_posEnum )
	throws SDMException {
		assert !(p_targetSpo == null && p_targetObject == null );
		
		if ( p_objects.isEmpty() ) {
			return false;
		}
	
		final Iterator<Object> iterator = p_objects.iterator();
	
		switch ( p_posEnum ) {
			case FIRST: {
				final Object targetInstanceObject = iterator.next();
				
				if ( match( targetInstanceObject, p_targetSpo, p_targetObject ) ) {
					return targetInstanceObject;
				}
				
				return null;
			}
			case BETWEEN: {
				if ( p_objects.size() > 2 ) {
					Object targetInstanceObject = iterator.next();
	
					while ( iterator.hasNext() ) {
						targetInstanceObject = iterator.next();
						
						if ( iterator.hasNext() ) {
							if ( match( targetInstanceObject, p_targetSpo, p_targetObject ) ) {
								return targetInstanceObject;
							}
						}
					}
				}
				
				return null;
			}
			case LAST: {
				Object targetInstanceObject = iterator.next();
				
				while ( iterator.hasNext() ) {
					targetInstanceObject = iterator.next();
				}
				
				if ( match( targetInstanceObject, p_targetSpo, p_targetObject ) ) {
					return targetInstanceObject;
				}
				
				return null;
			}
			default:
				throw new IllegalArgumentException( "Unknown poition constraint " + p_posEnum );
		}
	}
	
	private boolean match( 	final Object p_targetInstanceObject,
							final AbstractStoryPatternObject p_targetSpo, 
							final Object p_targetObject2 )
	throws SDMException {
		//checkUnresolvedProxy( p_targetInstanceObject );
		
		if ( p_targetObject2 == null ) {
			return patternMatcher.matchStoryPatternObject( p_targetSpo, p_targetInstanceObject );
		}

		//checkUnresolvedProxy( p_targetObject2 );
		
		return p_targetObject2 == p_targetInstanceObject;
	}
	// @Override
	// protected boolean checkLinkConstraints()
	// {
	// if (this.storyPatternLink.getLinkPositionConstraint() == null &&
	// this.storyPatternLink.getIncomingLinkOrderConstraints().isEmpty()
	// && this.storyPatternLink.getOutgoingLinkOrderConstraints().isEmpty())
	// {
	// return true;
	// }
	// else
	// {
	// assert this.storyPatternLink.getFeature().isMany();
	//
	// AbstractStoryPatternObject sourceSpo = this.storyPatternLink.getSource();
	// AbstractStoryPatternObject targetSpo = this.storyPatternLink.getTarget();
	//
	// assert this.patternMatcher.boundSPO.contains(sourceSpo);
	// assert this.patternMatcher.boundSPO.contains(targetSpo);
	//
	// Variable<EClassifier> sourceVariable =
	// this.patternMatcher.getVariablesScope().getVariable(sourceSpo.getName());
	//
	// assert sourceVariable != null;
	// assert sourceVariable.getValue() instanceof EObject;
	//
	// EObject sourceInstanceObject = (EObject) sourceVariable.getValue();
	//
	// Variable<EClassifier> targetVariable =
	// this.patternMatcher.getVariablesScope().getVariable(targetSpo.getName());
	//
	// assert targetVariable != null;
	// assert targetVariable.getValue() instanceof EObject;
	//
	// EObject targetInstanceObject = (EObject) targetVariable.getValue();
	//
	// @SuppressWarnings("unchecked")
	// List<Object> targetObjects = (List<Object>)
	// sourceInstanceObject.eGet(this.storyPatternLink.getFeature());
	//
	// /*
	// * If there is a link position constraint, check that the target
	// * instance object is at the specified position.
	// */
	// if (this.storyPatternLink.getLinkPositionConstraint() != null)
	// {
	// switch
	// (this.storyPatternLink.getLinkPositionConstraint().getConstraintType())
	// {
	// case FIRST:
	// {
	// if (targetObjects.get(0) != targetInstanceObject)
	// {
	// return false;
	// }
	//
	// break;
	// }
	//
	// case LAST:
	// {
	// if (targetObjects.get(targetObjects.size() - 1) != targetInstanceObject)
	// {
	// return false;
	// }
	//
	// break;
	// }
	//
	// default:
	// throw new UnsupportedOperationException();
	// }
	// }
	//
	// if (!this.storyPatternLink.getIncomingLinkOrderConstraints().isEmpty()
	// && !this.storyPatternLink.getOutgoingLinkOrderConstraints().isEmpty())
	// {
	// int targetIndex = targetObjects.indexOf(targetInstanceObject);
	//
	// /*
	// * Check all incoming link order constraints coming in from a
	// * link with a bound target object.
	// */
	// for (LinkOrderConstraint loc :
	// this.storyPatternLink.getIncomingLinkOrderConstraints())
	// {
	// if
	// (this.patternMatcher.boundSPO.contains(loc.getPredecessorLink().getTarget()))
	// {
	// Variable<EClassifier> variable =
	// this.patternMatcher.getVariablesScope().getVariable(
	// loc.getPredecessorLink().getTarget().getName());
	//
	// assert variable != null;
	// assert variable.getValue() != null;
	//
	// switch (loc.getConstraintType())
	// {
	// case DIRECT_SUCCESSOR:
	// {
	// if (targetIndex == 0 || targetObjects.get(targetIndex - 1) !=
	// variable.getValue())
	// {
	// return false;
	// }
	//
	// break;
	// }
	// case SUCCESSOR:
	// {
	// if (targetObjects.indexOf(variable.getValue()) >= targetIndex)
	// {
	// return false;
	// }
	//
	// break;
	// }
	// default:
	// throw new UnsupportedOperationException();
	// }
	// }
	// }
	//
	// /*
	// * Check all outgoing link order constraints going out to a link
	// * with a bound target object.
	// */
	// for (LinkOrderConstraint loc :
	// this.storyPatternLink.getOutgoingLinkOrderConstraints())
	// {
	// if
	// (this.patternMatcher.boundSPO.contains(loc.getSuccessorLink().getTarget()))
	// {
	// Variable<EClassifier> variable =
	// this.patternMatcher.getVariablesScope().getVariable(
	// loc.getSuccessorLink().getTarget().getName());
	//
	// assert variable != null;
	// assert variable.getValue() != null;
	//
	// switch (loc.getConstraintType())
	// {
	// case DIRECT_SUCCESSOR:
	// {
	// if (targetIndex == 0 || targetObjects.get(targetIndex + 1) !=
	// variable.getValue())
	// {
	// return false;
	// }
	//
	// break;
	// }
	// case SUCCESSOR:
	// {
	// if (targetObjects.indexOf(variable.getValue()) <= targetIndex)
	// {
	// return false;
	// }
	//
	// break;
	// }
	// default:
	// throw new UnsupportedOperationException();
	// }
	// }
	// }
	// }
	//
	// return true;
	// }
	// }
	
	private LinkPositionConstraintEnum positionConstraint( final Object p_object ) {
		final Variable<?> var = patternMatcher.getVariablesScope().getVariableFromValue( p_object );
		
		if ( var != null ) {
			final String spoName = var.getName();
			final EStructuralFeature feature = getStoryPatternLink().getFeature();
			final Enumeration<MatchTransaction> transactions = patternMatcher.matchingStack.elements();
			
			while ( transactions.hasMoreElements() ) {
				final MatchTransaction transaction = transactions.nextElement();
				
				if ( transaction instanceof CheckPatternPartTransaction<?, ?, ?, ?, ?> ) {
					final CheckPatternPartTransaction<?, ?, ?, ?, ?> checkTrans = (CheckPatternPartTransaction<?, ?, ?, ?, ?>) transaction;
					@SuppressWarnings("unchecked")
					final PatternPart<AbstractStoryPatternObject, ?, ?, ?, ?> patternPart = (PatternPart<AbstractStoryPatternObject, ?, ?, ?, ?>) checkTrans.getPatternPart();
					
					for ( final AbstractStoryPatternObject spo : patternPart.getStoryPatternObjects() ) {
						if ( spoName.equals( spo.getName() ) ) {
							for ( final AbstractStoryPatternLink link : spo.getIncomingLinks() ) {
								final StoryPatternLink spoLink = (StoryPatternLink) link;
								
								if ( feature == spoLink.getFeature() && spoLink.getLinkPositionConstraint() != null ) {
									return spoLink.getLinkPositionConstraint();
								}
							}
						}
					}
				}
			}
		}
		
		return null;
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void doCreateLink()
	{
		final EObject sourceEObject = (EObject) this.patternMatcher.getInstanceObject(this.storyPatternLink.getSource());
		final Object targetObject = this.patternMatcher.getInstanceObject(this.storyPatternLink.getTarget());

		if (!this.storyPatternLink.getFeature().isMany())
		{
			assert this.storyPatternLink.getLinkPositionConstraint() == LinkPositionConstraintEnum.UNCONSTRAINED;

			/*
			 * Do not create instance links for optional create links that
			 * already exist.
			 */
			if (this.storyPatternLink.getMatchType() != MatchTypeEnum.OPTIONAL
					|| sourceEObject.eGet(this.storyPatternLink.getFeature()) != targetObject)
			{
				sourceEObject.eSet(this.storyPatternLink.getFeature(), targetObject);
			}
		}
		else
		{
			final List<Object> objects = (List<Object>) sourceEObject.eGet(this.storyPatternLink.getFeature());

			/*
			 * Do not create instance links for optional create link that
			 * already exist.
			 */
			if ( storyPatternLink.getMatchType() != MatchTypeEnum.OPTIONAL || !objects.contains( targetObject ) ) {
				// If there is a link position constraint, insert the element at the front or end of the list
				if ( storyPatternLink.getLinkPositionConstraint() != null ) {
					switch ( storyPatternLink.getLinkPositionConstraint() ) {
						case UNCONSTRAINED:
						case LAST: {
							objects.add( targetObject );
							
							break;
						}
						case FIRST:	{
							objects.add( 0, targetObject );
							
							break;
						}
						case BETWEEN: {
							if ( objects.isEmpty() ) {
								objects.add( targetObject );
							}
							else if ( objects.size() == 1 ) {
								// Determine if this is the first or last element
								final LinkPositionConstraintEnum positionConstraint = positionConstraint( objects.get( 0 ) );
								
								if ( positionConstraint == LinkPositionConstraintEnum.LAST ) {
									objects.add( 0, targetObject );
								}
								else {
									objects.add( targetObject );
								}
							}
							else {
								objects.add( 1, targetObject );
							}
							
							break;
						}
					}
				}
				else {
					boolean done = false;

					/*
					 * If there are incoming link order constraints, pick one
					 * that is connected to a link with an existing target
					 * object and insert the current target object after it.
					 */
					for (final LinkOrderConstraint loc : this.storyPatternLink.getIncomingLinkOrderConstraints())
					{
						assert loc.getConstraintType() == LinkOrderConstraintEnum.DIRECT_SUCCESSOR;

						final Variable<EClassifier> variable = this.patternMatcher.getVariablesScope().getVariable(
								loc.getPredecessorLink().getTarget().getName());

						if (variable != null)
						{
							objects.add(objects.indexOf(variable.getValue()) + 1, targetObject);

							done = true;

							break;
						}
					}

					if (!done)
					{
						/*
						 * If there are outgoing link order constraints, pick
						 * one that is connected to a link with an existing
						 * target object and insert the current target object
						 * before it.
						 */
						for (final LinkOrderConstraint loc : this.storyPatternLink.getOutgoingLinkOrderConstraints())
						{
							assert loc.getConstraintType() == LinkOrderConstraintEnum.DIRECT_SUCCESSOR;

							final Variable<EClassifier> variable = this.patternMatcher.getVariablesScope().getVariable(
									loc.getSuccessorLink().getTarget().getName());

							if (variable != null)
							{
								objects.add(objects.indexOf(variable.getValue()), targetObject);

								done = true;

								break;
							}
						}
					}

					if (!done) {
						/*
						 * Simply insert the target object at the end.
						 */
						objects.add(targetObject);
					}
				}
				/*
				 * Simply insert the target object at the end.
				 */
				//objects.add(targetObject);
			}
		}

		this.patternMatcher.getNotificationEmitter().instanceLinkCreated(this.storyPatternLink.getSource(), sourceEObject,
				this.storyPatternLink, this.storyPatternLink.getTarget(), targetObject, this.patternMatcher.getVariablesScope(),
				this.patternMatcher);
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void doDestroyLink(final Map<AbstractStoryPatternObject, Object> deletedObjects)
	{
		assert this.storyPatternLink.getModifier() == ModifierEnum.DESTROY;

		final AbstractStoryPatternObject sourceSpo = this.storyPatternLink.getSource();
		final AbstractStoryPatternObject targetSpo = this.storyPatternLink.getTarget();

		assert deletedObjects.get(sourceSpo) == null || deletedObjects.get(sourceSpo) instanceof EObject;

		EObject sourceInstanceObject = (EObject) deletedObjects.get(sourceSpo);

		Object targetInstanceObject = deletedObjects.get(targetSpo);

		if (sourceInstanceObject == null)
		{
			/*
			 * The source spo was not destroyed
			 */
			final Variable<EClassifier> sourceVariable = this.patternMatcher.getVariablesScope().getVariable(sourceSpo.getName());

			if (sourceVariable != null)
			{
				assert sourceSpo.getMatchType() != MatchTypeEnum.OPTIONAL;
				assert sourceVariable.getValue() instanceof EObject;

				sourceInstanceObject = (EObject) sourceVariable.getValue();
			}
		}

		if (targetInstanceObject == null)
		{
			/*
			 * The target spo was not destroyed.
			 */
			final Variable<EClassifier> targetVariable = this.patternMatcher.getVariablesScope().getVariable(targetSpo.getName());

			if (targetVariable != null)
			{
				assert targetSpo.getMatchType() != MatchTypeEnum.OPTIONAL;

				targetInstanceObject = targetVariable.getValue();
			}
		}

		if (sourceInstanceObject != null && targetInstanceObject != null)
		{
			/*
			 * Delete the link
			 */
			if (!this.storyPatternLink.getFeature().isMany())
			{
				sourceInstanceObject.eSet(this.storyPatternLink.getFeature(), null);
			}
			else
			{
				((Collection<Object>) sourceInstanceObject.eGet(this.storyPatternLink.getFeature())).remove(targetInstanceObject);
			}

			this.patternMatcher.getNotificationEmitter().instanceLinkDestroyed(sourceSpo, sourceInstanceObject, this.storyPatternLink,
					targetSpo, targetInstanceObject, this.patternMatcher.getVariablesScope(), this.patternMatcher);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public int calculateMatchingCost()
	{
		final AbstractStoryPatternObject sourceSpo = this.storyPatternLink.getSource();
		final AbstractStoryPatternObject targetSpo = this.storyPatternLink.getTarget();

		assert !(this.patternMatcher.isBound(sourceSpo) && this.patternMatcher.isBound(targetSpo));

		final EStructuralFeature eStructuralFeature = this.storyPatternLink.getFeature();

		assert eStructuralFeature != null;

		if (this.patternMatcher.isBound(sourceSpo))
		{
			if (!eStructuralFeature.isMany())
			{
				return 1 + this.getNumberOfIncomingUnidirectionalLinks(targetSpo);
			}
			else
			{
				final Variable<EClassifier> sourceVariable = this.patternMatcher.getVariablesScope().getVariable(sourceSpo.getName());

				assert sourceVariable.getValue() instanceof EObject;

				final EObject sourceEObject = (EObject) sourceVariable.getValue();

				return ((Collection<Object>) sourceEObject.eGet(eStructuralFeature)).size()
						+ this.storyPatternLink.getIncomingLinkOrderConstraints().size()
						+ this.getNumberOfIncomingUnidirectionalLinks(targetSpo);
			}
		}
		else if (this.patternMatcher.isBound(targetSpo))
		{
			if (eStructuralFeature instanceof EReference)
			{
				final EReference eReference = (EReference) eStructuralFeature;

				if (eReference.isContainment())
				{
					/*
					 * We can get the direct container by calling eContainer()
					 * on the target object.
					 */
					return 2 + this.storyPatternLink.getIncomingLinkOrderConstraints().size()
							+ this.getNumberOfIncomingUnidirectionalLinks(sourceSpo);
				}
				else if (eReference.getEOpposite() != null)
				{
					if (!eReference.getEOpposite().isMany() || this.storyPatternLink.getLinkPositionConstraint() != null)
					{
						return 1 + this.getNumberOfIncomingUnidirectionalLinks(sourceSpo);
					}
					else
					{
						final Variable<EClassifier> targetVariable = this.patternMatcher.getVariablesScope().getVariable(
								targetSpo.getName());

						assert targetVariable.getValue() instanceof EObject;

						final EObject targetEObject = (EObject) targetVariable.getValue();

						return ((Collection<EObject>) targetEObject.eGet(eReference.getEOpposite())).size()
								+ this.storyPatternLink.getIncomingLinkOrderConstraints().size()
								+ this.getNumberOfIncomingUnidirectionalLinks(sourceSpo);
					}
				}
				else
				{
					return PatternPart.MATCHING_NOT_POSSIBLE;
				}
			}
			else
			{
				return PatternPart.MATCHING_NOT_POSSIBLE;
			}
		}
		else
		{
			return PatternPart.MATCHING_NOT_POSSIBLE;
		}
	}

	@Override
	public EMatchType doGetMatchType()
	{
		switch (this.storyPatternLink.getModifier())
		{
			case NONE:
				switch (this.storyPatternLink.getMatchType())
				{
					case DEFAULT:
						return EMatchType.MANDATORY;
					case OPTIONAL:
						return EMatchType.OPTIONAL;
				}
			case CREATE:
				switch (this.storyPatternLink.getMatchType())
				{
					case DEFAULT:
						return EMatchType.OPTIONAL;
					case OPTIONAL:
						return EMatchType.OPTIONAL;
				}
			case DESTROY:
				switch (this.storyPatternLink.getMatchType())
				{
					case DEFAULT:
						return EMatchType.MANDATORY;
					case OPTIONAL:
						return EMatchType.OPTIONAL;
				}
			default:
				throw new UnsupportedOperationException();
		}
	}
}
